package com.extremelogic.common.jclap;

import java.util.Map;

import com.extremelogic.common.jclap.exception.ArgumentException;
import com.extremelogic.common.jclap.exception.ArgumentInvalidException;

/**
 * @author Jan So (extreme.logic at yahoo.com)
 */
public class Argument {

    /**
     * list of arguments being used by the program.
     */
    private ArgumentList list = null;

    /**
     * delimiter string with default value.
     */
    private String delimiter = "--";

    /**
     * argument input. This comes from Main.
     */
    private String[] inArgs;

    /**
     * Flag to determine if options should be shown
     * if an exception occurs.
     */
    private boolean showOptions = false;

    /**
     * Name of the main program.
     */
    private String programName;

    /**
     * Description of the main program.
     */
    private String programDescription;

    /**
     * Internal use only to hold string length of the option.
     */
    private int flagLength = 0;

    /**
     * Variable to determine if stack tract should be shown.
     */
    private boolean showStackTrace = false;

    /**
     * Variable to determine if an exception had occurred.
     */
    private boolean exceptionOccured = false;

    /**
     * Variable to determine if re-throwing an exception is allowed.
     */
    private boolean rethrowException = false;

    /**
     * Object that retrieves the arguments/options.
     */
    private ArgumentRetriever argumentRetriever = null;

    /**
     * List of argument value mapping.
     */
    private ArgumentValueMap valueList = null;

    /**
     * <p>Constructor with argument parameters. The argument being passed is the
     * one passed in the <code>main(String[] args)</code></p>
     * 
     * @param args Main arguments
     */
    public Argument(final String[] args) {
        StackTraceElement element = null;
        inArgs = args;

        element = Thread.currentThread().getStackTrace()[2];
        parseSimpleName(element.getClassName());

        list = new ArgumentList();
        valueList = new ArgumentValueMap();
    }

    /**
     * Initializes boolean typed options.
     */
    private void initBoolean() {
        for (Map.Entry<String, ArgumentWrapper> entry : list.entrySet()) {
            String key = entry.getKey();
            ArgumentWrapper value = entry.getValue();

            if (0 == value.getParamCount()) {
                valueList.put(key, Boolean.FALSE);
            }
        }
    }

    /**
     * @param className fully qualified class name
     */
    private void parseSimpleName(final String className) {
        int index;
        String simpleName;

        index = className.lastIndexOf(".");
        simpleName = className.substring(index + 1);

        setProgramName(simpleName);
    }

    /**
     * <p>Displays the usage or help options.</p>
     */
    public final void displayOptions() {
        String progDesc;

        progDesc = getProgramDescription();

        System.out.println();
        System.out.println("Usage: " + getProgramName() + " [options]");

        if (null != progDesc && progDesc.trim().length() > 0) {
            System.out.println(getProgramDescription());
        }
        System.out.println();

        for (ArgumentWrapper flags : list.values()) {
            String format;
            String delim;
            String flag;
            String desc;

            format = "  %s%-" + (getFlagLength()) + "s %s\n";
            delim = getDelimiter();
            flag = flags.getFlag();
            desc = flags.getDescription();

            System.out.format(format, delim, flag, desc);
        }
    }

    /**
     * <p>Processes the arguments for later argument retrieval.</p>
     *
     * @throws ArgumentException if an exception occurs and is
     * allowed to be re-thrown
     */
    public final void processArguments() throws ArgumentException {
        ArgumentRetriever ar;

        try {
            validate();
            ar = new ArgumentRetriever(getValueList());
            setArgumentRetriever(ar);
        } catch (ArgumentException e) {

            setExceptionOccured(true);

            if (isShowStackTrace()) {
                e.printStackTrace();
            }

            if (isShowOptions()) {
                displayOptions();
            }

            if (isRethrowException()) {
                throw e;
            }
        }
    }

    /**
     * @throws ArgumentException if there is an error in validating the inputs
     */
    private void validate() throws ArgumentException {
        boolean display = false;

        if (0 == inArgs.length) {
            display = true;
        }

        initBoolean();

        // validate argument input syntax
        for (int i = 0; i < inArgs.length; i++) {
            ArgumentFlagWrapper afw = null;

            afw = new ArgumentFlagWrapper(delimiter, inArgs[i]);

            if (afw.hasDelimiter()) {
                if (list.containsFlag(afw.getFlag())) {
                    ArgumentWrapper aq;
                    int paramCount;
                    aq = list.get(afw.getFlag());
                    paramCount = aq.getParamCount();

                    if ((paramCount + i) > inArgs.length) {
                        throw new ArgumentInvalidException("parameter count",
                        "" + inArgs.length, "" + paramCount + inArgs.length);
                    } else {

                        if (0 == paramCount) {
                            if ((getDelimiter() + aq.getFlag())
                                .equals(inArgs[i])) {
                                valueList.put(afw.getFlag(), Boolean.TRUE);
                            }
                        } else {
                            String[] s = new String[paramCount];
                            for (int j = 1; j < paramCount + 1; j++) {
                                String tmp;

                                if ((j + i) >= inArgs.length) {
                                    throw new ArgumentInvalidException(
                                        "sub parameter count", "" + j, ""
                                            + (j - 1));
                                }

                                tmp = inArgs[j + i];
                                s[j - 1] = tmp;

                                if (tmp.contains(delimiter)) {
                                    throw new ArgumentInvalidException(
                                        ". Parameter missing ", afw.getFlag());
                                }
                            }
                            valueList.put(afw.getFlag(), s);
                        }
                        i += (paramCount);
                    }
                } else {
                    throw new ArgumentInvalidException("flag", inArgs[i]);
                    // display = true;
                }
            } else {
                throw new ArgumentInvalidException("delimiter", inArgs[i]);
            }
        }

        if (isShowOptions() && display) {
            displayOptions();
        }
    }

    /**
     * <p>Method to add arguments. Value being passed also is being
     * used by the <code>displayOptions()</code> method.</p>
     * 
     * @param option flag used in the arguments
     * @param description the description of the option provided
     * @param paramCount parameter count to be used by the option
     */
    public final void addArgument(final String option,
           final String description, final int paramCount) {
        ArgumentWrapper aw = null;
        int len;

        aw = new ArgumentWrapper();

        aw.setFlag(option);
        aw.setDescription(description);
        aw.setParamCount(paramCount);

        list.put(option, aw);

        len = option.length();

        if (getFlagLength() < len) {
            setFlagLength(len);
        }
    }

    /**
     * <p>Method to add arguments. Value being passed also is being
     * used by the <code>displayOptions()</code> method.</p>
     * 
     * @param option flag used in the arguments
     * @param description the description of the option provided
     */
    public final void addBooleanArgument(final String option,
        final String description) {
        addArgument(option, description, 0);
    }

    /**
     * @return the delimiter
     */
    protected final String getDelimiter() {
        return delimiter;
    }

    /**
     * <p>The delimiter being used to parse the options.
     * Default delimiter is "--" e.g --ls</p>
     * 
     * @param delim the delimiter to be used by the option
     */
    public final void setDelimiter(final String delim) {
        this.delimiter = delim;
    }

    /**
     * @return <code>true</code> if help options is to be displayed;
     * else <code>false</code>
     */
    private boolean isShowOptions() {
        return showOptions;
    }

    /**
     * <p>Method that allows the user to allow options to be displayed if
     * an error occurs.</p>
     * 
     * @param so set option to <code>true</code> if its to be shown
     * automatically if an exception occurs else <code>false</code>
     */
    public final void setShowOptions(final boolean so) {
        this.showOptions = so;
    }

    /**
     * @return program name
     */
    private String getProgramName() {
        return programName;
    }

    /**
     * @param pn the class name which executed
     */
    private void setProgramName(final String pn) {
        this.programName = pn;
    }

    /**
     * @return program description
     */
    private String getProgramDescription() {
        return programDescription;
    }

    /**
     * <p>Describes the calling class.</p>
     * 
     * @param pd description of the program
     */
    public final void setProgramDescription(final String pd) {
        this.programDescription = pd;
    }

    /**
     * @return length of the option
     */
    private int getFlagLength() {
        return flagLength;
    }

    /**
     * @param fl length of the option
     */
    private void setFlagLength(final int fl) {
        this.flagLength = fl;
    }

    /**
     * @return list of argument option objects
     */
    private ArgumentValueMap getValueList() {
        return valueList;
    }

    /**
     * <p>Checks if showing of stack traces is allowed by the user.</p>
     * 
     * @return <code>true</code> if a stack trace is allowed to
     *  be shown otherwise <code>false</code>
     */
    public final boolean isShowStackTrace() {
        return showStackTrace;
    }

    /**
     * <p>Allows stack trace to be shown when an exception occurs.</p>
     *
     * @param show <code>true</code> if stack trace should be shown
     * automatically else set to <code>false</code>
     */
    public final void setShowStackTrace(final boolean show) {
        this.showStackTrace = show;
    }

    /**
     * <p>Allows user to check if an exception occurred while parsing
     * the arguments</p>
     * 
     * @return <code>true</code> if an exception occurred
     *  otherwise <code>false</code>
     */
    public final boolean isExceptionOccured() {
        return exceptionOccured;
    }

    /**
     * <p>Sets value to distinguish that an exception had occurred.</p>
     *
     * @param exception <code>true</code> if an exception had occurred
     *  else set to <code>false</code>
     */
    private void setExceptionOccured(final boolean exception) {
        this.exceptionOccured = exception;
    }

    /**
     * <p>Allows user to check if re-throwing of an exception is set.</p>
     * 
     * @return <code>true</code> if exception is re-thrown
     *  else <code>false</code>
     */
    public final boolean isRethrowException() {
        return rethrowException;
    }

    /**
     * <p>Sets value to re-throw an exception.</p>
     *
     * @param rethrow <code>true</code> if the exception is to be re-thrown 
     * to caller else set to <code>false</code>
     */
    public final void setRethrowException(final boolean rethrow) {
        this.rethrowException = rethrow;
    }

    /**
     * <p>Method to retrieve values.</p>
     * 
     * @return <code>ArgumentRetriever</code> object
     *  that retrieves argument values
     */
    public final ArgumentRetriever getArgumentRetriever() {
        return argumentRetriever;
    }

    /**
     * @param argRetriever Populated argument objects
     */
    private void setArgumentRetriever(final ArgumentRetriever argRetriever) {
        this.argumentRetriever = argRetriever;
    }
}
